from turbot.blocks import (
    section_block, text_block, button_block, actions_block, divider_block
)
from turbot.hunt import find_hunt_for_hunt_id
from turbot.puzzle import (
    puzzle_channel_topic, puzzle_channel_description,
)
from turbot.sheets import (
    sheets_create, sheets_create_for_puzzle, sheets_create_folder
)
from turbot.slack import slack_send_message, slack_channel_members
from turbot.channel import channel_url
from boto3.dynamodb.conditions import Key

TURBOT_USER_ID = 'U01B9QM4P9R'

# This is a list of user IDs that should not be auto-invited to new channels
AUTO_INVITE_DENY_LIST = (
    'U01B9QM4P9R',  # Turbot itself
    'U01HQABASLW',  # Lavi
)

events = {}

lambda_success = {'statusCode': 200}
lambda_error = {'statusCode': 400}

def hunt_link_block(turb, hunt):

    name = hunt['name']
    channel_id = hunt['channel_id']

    hunt_link = "*<{}|{}>*".format(channel_url(channel_id), name)

    return section_block(text_block(hunt_link))

def home(turb, user_id):
    """Returns a view to be published as the turbot home tab for user_id

    The return value is a dictionary suitable to be published to the
    Slack views_publish API."""

    # Behave cleanly if there is no "turbot" table at all yet.
    try:
        response = turb.table.scan(
            IndexName="is_hunt_index",
        )
        hunts = response['Items']
    except Exception:
        hunts = []

    my_hunt_blocks = []
    available_hunt_blocks = []
    for hunt in hunts:
        if not hunt['active']:
            continue
        if user_id in slack_channel_members(turb.slack_client,
                                            hunt['channel_id']):
            my_hunt_blocks.append(hunt_link_block(turb, hunt))
        else:
            available_hunt_blocks.append(hunt_link_block(turb, hunt))

    if len(my_hunt_blocks):
        my_hunt_blocks = [
            section_block(text_block("*Hunts you belong to:*")),
            divider_block(),
            * my_hunt_blocks
        ]
    else:
        my_hunt_blocks.append([
            section_block(text_block("You do not belong to any hunts"))
        ])

    if len(available_hunt_blocks):
        available_hunt_blocks = [
            section_block(text_block("*Hunts you can join:*")),
            divider_block(),
            * available_hunt_blocks
        ]

    return {
        "type": "home",
        "blocks": [
            * my_hunt_blocks,
            * available_hunt_blocks,
            actions_block(button_block("New hunt", "new_hunt"))
        ]
    }

def app_home_opened(turb, event):
    """Handler for the app_home_opened event

    This event occurs when a user visits the home tab for the turbot app.
    In response to this event we need to publish a view for the user."""

    user_id = event['user']
    view = home(turb, user_id)
    turb.slack_client.views_publish(user_id=user_id, view=view)
    return lambda_success

events['app_home_opened'] = app_home_opened

def hunt_channel_created(turb, channel_name, channel_id):
    """Creates a Google sheet for a newly-created hunt channel"""

    # First see if we can find an entry for this hunt in the database.
    # If not, simply return an error and let Slack retry
    response = turb.table.query(
        IndexName='channel_id_index',
        KeyConditionExpression=Key("channel_id").eq(channel_id)
    )
    if 'Items' not in response:
        print("Warning: Cannot find channel_id {} in turbot table. "
              .format(channel_id) + "Letting Slack retry this event")
        return lambda_error

    hunt = response['Items'][0]

    if 'sheet_url' in hunt:
        print("Info: channel_id {} already has sheet_url {}. Exiting."
              .format(channel_id, hunt['sheet_url']))
        return lambda_success

    # Before launching into sheet creation, indicate that we're doing this
    # in the database. This way, if we take too long to create the sheet
    # and Slack retries the event, that next event will see this 'pending'
    # string and cleanly return (eliminating all future retries).
    hunt['sheet_url'] = 'pending'
    turb.table.put_item(Item=hunt)

    # Also, let the channel users know what we are up to
    slack_send_message(
        turb.slack_client, channel_id,
        "Welcome to the channel for the {} hunt! ".format(hunt['name'])
        + "Please wait a moment or two while I create some backend resources.")

    # Create a new folder within Google drive for the hunt in two parts:

    # ... first, a top-level folder named for the hunt
    hunt_folder = sheets_create_folder(turb, hunt['hunt_id'])

    # ... second, a folder within that named "turbot"
    #
    # The concept here is that non-turbot related content could
    # be placed adjacent to the turbot folder.
    hunt['folder_id'] = sheets_create_folder(turb, "turbot",
                                             parents = [hunt_folder])

    # Create a sheet for the hunt
    sheet = sheets_create(turb, hunt['name'], hunt['folder_id'])
    hunt['sheet_url'] = sheet['url']

    # Message the channel with the URL of the sheet
    slack_send_message(turb.slack_client, channel_id,
                       "Sheet created for this hunt: {}".format(sheet['url']))

    # Mark the hunt as active now
    hunt['active'] = True

    # Update the database with all the changes we have made to the hunt
    turb.table.put_item(Item=hunt)

    # Message the hunt channel that the database is ready
    slack_send_message(
        turb.slack_client, channel_id,
        "Thank you for waiting. This hunt is now ready to begin! "
        + "Type `/new` to create a puzzle for the hunt and `/help` for help.")

    return lambda_success

def set_channel_topic_and_description(turb, puzzle):

    channel_id = puzzle['channel_id']

    topic = puzzle_channel_topic(puzzle)
    description = puzzle_channel_description(puzzle)

    turb.slack_client.conversations_setPurpose(channel=channel_id,
                                               purpose=description)
    turb.slack_client.conversations_setTopic(channel=channel_id,
                                             topic=topic)

def puzzle_channel_created(turb, channel_name, channel_id):
    """Creates sheet and invites user for a newly-created puzzle channel"""

    # First see if we can find an entry for this puzzle in the database.
    # If not, simply return an error and let Slack retry
    response = turb.table.query(
        IndexName="channel_id_index",
        KeyConditionExpression=Key("channel_id").eq(channel_id),
    )
    if 'Items' not in response:
        print("Warning: Cannot find channel_id {} in turbot table. "
              .format(channel_id) + "Letting Slack retry this event")
        return lambda_error

    puzzle = response['Items'][0]

    if 'sheet_url' in puzzle:
        print("Info: channel_id {} already has sheet_url {}. Exiting."
              .format(channel_id, puzzle['sheet_url']))
        return lambda_success

    # We need hunt from the database to know which folder to create
    # the sheet in.
    hunt_id = puzzle['hunt_id']
    hunt = find_hunt_for_hunt_id(turb, hunt_id)

    # Before launching into sheet creation, indicate that we're doing this
    # in the database. This way, if we take too long to create the sheet
    # and Slack retries the event, that next event will see this 'pending'
    # string and cleanly return (eliminating all future retries).
    puzzle['sheet_url'] = 'pending'
    puzzle['channel_url'] = channel_url(channel_id)
    turb.table.put_item(Item=puzzle)

    # Create a sheet for the puzzle
    sheet = sheets_create_for_puzzle(turb, puzzle, hunt['folder_id'])

    # Update the database with the URL of the sheet
    puzzle['sheet_url'] = sheet['url']
    turb.table.put_item(Item=puzzle)

    # Get the new sheet_url into the channel topic and description
    set_channel_topic_and_description(turb, puzzle)

    # Lookup and invite all users from this hunt to this new puzzle

    # Find all members of the hunt channel
    members = slack_channel_members(turb.slack_client, hunt['channel_id'])

    # Filter out Turbot's own ID as well as that of any user who
    # has opted out of being auto-invited
    members = [m for m in members if m not in AUTO_INVITE_DENY_LIST]

    ### With the code below disabling auto-invitation, be honest about that.
    slack_send_message(
        turb.slack_client, channel_id,
        "Not inviting all members from the hunt channel: "
        + "<#{}>".format(hunt['channel_id']))

    # Invite those members to the puzzle channel (in chunks of 500)
    cursor = 0

    ### This line of code disables the auto-invitation of hunt members
    ### to channels, (which I'm throwing in at the request of Avram
    ### who is mostly-solo solving puzzles after Mystery Hunt 2022 is
    ### over).
    cursor = len(members)

    while cursor < len(members):
        turb.slack_client.conversations_invite(
            channel=channel_id,
            users=members[cursor:cursor + 500])
        cursor += 500

    # And finally, give a welcome message with some documentation
    # on how to update the state of the puzzle in the database.
    welcome_msg = (
        "Welcome! This channel is the primary place to "
        + "discuss things as the team works together to solve the "
        + "puzzle \"{}\". ".format(puzzle['name'])
    )

    if 'url' in puzzle:
        welcome_msg += (
            "See the <{}|puzzle itself> ".format(puzzle['url'])
            + "for what was originally presented to us. "
        )

    welcome_msg += (
        "Actual puzzle solving work will take place within the following " +
        "<{}|shared spreadsheet> ".format(puzzle['sheet_url']) +
        "\n\n"
        "Common commands for updating the puzzle are `/state NEW STATE`, " +
        "`/tag NEW_TAG`, and `/solved SOLUTION` . See `/help` for details " +
        "and for additional commands."
    )

    turb.slack_client.chat_postMessage(channel=channel_id, text=welcome_msg)

    # Finally, finally, notify the hunt channel about the new puzzle
    slack_send_message(
        turb.slack_client, hunt['channel_id'],
        "New puzzle available: <{}|{}>".format(
            puzzle['channel_url'],
            puzzle['name'])
    )

    return lambda_success

def channel_created(turb, event):

    channel = event['channel']
    channel_id = channel['id']
    channel_name = channel['name']
    creator = channel['creator']

    # Ignore any channels that turbot didn't create
    if creator != TURBOT_USER_ID:
        print("channel_created: Not a turbot-created channel. Exiting.")
        return lambda_success

    # The presence of a hyphen determines whether this is a puzzle
    # channel or a hunt channel.
    if '-' in channel_name:
        return puzzle_channel_created(turb, channel_name, channel_id)
    else:
        return hunt_channel_created(turb, channel_name, channel_id)

events['channel_created'] = channel_created
